# flcheck --
#
# This file implements package flcheck, a fusion logic modelchecker 
# using the CUDD BDD package
#
# Copyright (C) 2009, 2010  Antonio Cau

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Initialise the BuDDy BDD 

proc Init-data-manager {} {
variable count
variable pair 
set count 0
bdd_init 1000 100
bdd_autoreorder $bdd_tcl::BDD_REORDER_SIFTITE
set pair [bdd_newpair]
bdd_gbc_quiet
}

# Terminate the BuDDy BDD
#
proc Quit-data-manager {} {
 variable pair 
 variable count
 
 bdd_freepair $pair
 bdd_done
}

# Creates a new BDD variable
#
proc bddVar0 {} {
  variable count
  #puts "count: $count"
  set node [bdd_ithvar $count]
  incr count
  bdd_addref $node
  return $node
}

# Creates a new BDD variable
# and assigns it to node
proc bddVar {node} {
  variable count
  upvar 1 $node var
  set var [bdd_ithvar $count]
  incr count
  bdd_addref $var
}

# One node BDD 
#
proc One {} {
    return [bdd_addref [bdd_true]]
}

# Zero node BDD
#
proc Zero {} {
    return [bdd_addref  [bdd_false]]
}

# Computes the NAND of two BDDs a and b
#
proc Nand {a b} {
    return [bdd_addref [bdd_apply $a $b $bdd_tcl::bddop_nand]]
}

# Computes the AND of two BDDs a and b
#
proc And {a b} {
  return [bdd_addref [bdd_apply $a $b $bdd_tcl::bddop_and]]
}

# Computes the OR of two BDDs a and b
#
proc Or {a b} {
  return [bdd_addref [bdd_apply $a $b $bdd_tcl::bddop_or]]
}

# Computes the XOR of two BDDs a and b
#
proc Xor {a b} {
  return [bdd_addref [bdd_xor $a $b]]
}

# Computes the XNOR of two BDDs a and b
#
proc Xnor {a b} {
  return [bdd_addref [bdd_apply $a $b $bdd_tcl::bddop_biimp]]
}

# Computes the NOR of two BDDs a and b
#
proc Nor {a b} {
  return [bdd_addref [bdd_apply $a $b $bdd_tcl::bddop_nor]]
}

# Computes the NOT of two BDDs a and b
#
proc Not {a} {
  return [bdd_addref [bdd_not $a]]
}

# Returns 1 if BDD a is less than or equal to b
#
proc Leq {a b} {
 set tmp [bdd_addref [bdd_restrict $b $a]]
    if { $tmp == [One] } {
      bdd_delref $tmp
      return 1
    } else {
      bdd_delref $tmp
      return 0
    }
}

# Computes the IMP of two BDDs a and b
#
proc Imp {a b} {
    return [bdd_addref [bdd_apply $a $b $bdd_tcl::bddop_imp]]
}

# Computes the EQV of two BDDs a and b
#
proc Eqv {a b} {
    return [Xnor $a $b] 
}

#Existentially abstracts all the variables b from a
#
proc Exists {a b} {
    return [bdd_addref  [bdd_exist $b $a] ]
}

# Takes the AND of two BDDs and simultaneously abstracts the variables in cube
#
proc AndExists {cube f g} {
    return [bdd_addref  [bdd_appex $f $g  $bdd_tcl::bddop_and $cube] ]
}
proc Swap {f pair} {    
    return [bdd_addref [bdd_replace $f $pair]]
}

#proc SwapVariables {f a b} {
#    set pair [bdd_newpair]
#    if {[llength $a] != [llength $b]} {
#     insert "SwapVariables: given variable sets of different size" red
#    }
#    set len [llength $a]
#    puts "len : $len"
#    bdd_printorder
#    for {set i 0} {$i<$len} {incr i} {
#        puts $pair
#        puts "i: $i"
#        puts "a:"
#        puts "b:"
#        bdd_setbddpair $pair [lindex $a $i] [lindex $b $i]
#        
#    }
#    
#    set node [bdd_addref [bdd_replace $f $pair]]
#    bdd_resetpair $pair
#    bdd_freepair $pair
#    return $node
#}

# Computes the AND of a list of BDDs
#
proc RAnd {name alist} {
    upvar $name x
    if {![info exist x]} {
      set x [One]
    }
    foreach e $alist {
        set x [And $x $e]
    }
}

# Computes the OR of a list of BDDs
#
proc ROr {name alist} {
    upvar $name x
    if {![info exist x]} {
      set x [Zero]
    }
    foreach e $alist {
        set x [Or $x $e]
    }
}


#proc ** {x args} {
#    foreach e $args {
#        set x [And $x $e]
#    }
#    return $x
#}

#proc ++ {x args} {
#    foreach e $args {
#        set x [Or $x $e]
#    }
#    return $x
#}


# defines * to be an abbreviation for ::bdd_tcl::And 
interp alias {} * {} ::bdd_tcl::And
# defines + to be an abbreviation for ::bdd_tcl::Or 
interp alias {} + {} ::bdd_tcl::Or
# defines ! to be an abbreviation for ::bdd_tcl::Not
interp alias {} ! {} ::bdd_tcl::Not
# defines ^ to be an abbreviation for ::bdd_tcl::Xor
interp alias {} ^ {} ::bdd_tcl::Xor
# defines Eqv to be an abbreviation for ::bdd_tcl::Eqv
interp alias {} Eqv {} ::bdd_tcl::Eqv

# BDD routines
#

# Initialise the BDD of normal and primed  (in)dependent variables  
#
proc init_my_basic_vars {suffix_name_list} {
   variable a {}
   variable b {}
   variable manager
   variable my_names
   variable a_cube [One]
   variable b_cube [One]
   variable pair

     
    set num_vars [llength $suffix_name_list]
    insert "number of variables $num_vars"
    bdd_setvarnum [expr 2*$num_vars]
    set my_names {}
    for {set j 0} {$j< $num_vars} {incr j} {
        set suffix_name [lindex $suffix_name_list $j]
        #set tmp1 [bddVar0]
        lappend a [bddVar0]
        #bdd_delref $tmp1
        #set tmp2 [bddVar0]
        lappend b [bddVar0]
        #bdd_delref $tmp2
        bdd_setpair $pair [expr 2*$j+1] [expr 2*$j] 

        #puts "pair: [expr 2*$j+1], [expr 2*$j]"
        #bdd_setbddpair $pair $tmp1 $tmp2
        lappend my_names "a_$suffix_name"
        lappend my_names "b_$suffix_name"
        
        set a_cube  [* $a_cube [lindex $a $j]]
        set b_cube  [* $b_cube [lindex $b $j]]
    }
    bdd_varblockall
    #insert "finished init variables"
}

# show the current number of live nodes 
#
proc check_node_stats {} { 
  
  insert "Buddy current number of live nodes=[bdd_getnodenum]" blue
}


# give a dot-based representation of the BDD in tab
#
proc dot {onodes} {
  variable manager
  global c tab bddfont nodisplay
  if {$nodisplay==0 } {
    #set leno [llength $onodes]
    #set tempo [new_node_array $leno]
    #for {set i 0} {$i<$leno} {incr i} {
    #    array_set_node $tempo $i [lindex $onodes $i]
    #}
    #set tempin [new_name_array [llength $iname]]
    #for {set i 0} {$i<[llength $iname]} {incr i} {
    #    array_set_name $tempin $i [lindex $iname $i]
    #}
    #set tempout [new_name_array [llength $oname]]
    #for {set i 0} {$i<[llength $oname]} {incr i} {
    #    array_set_name $tempout $i [lindex $oname $i]
    #}
    
    new_tab .top18 $tab
    incr tab
    #puts "tab; $tab"
    dotty $onodes 
    insert "start generating dot graph" blue
    if [string equal $::tcl_platform(platform) "windows"] { 
      set f [exec "c:/Program Files/Graphviz2.20/bin/dot.exe" -Ttk -Nstyle=filled < "./dotty.dot" > dotty.tcl]
    } else {
      set f [exec dot -Ttk -Nstyle=filled < "./dotty.dot" > dotty.tcl]
    }
    $c delete node edge
    source dotty.tcl   
    $c itemconfigure label -font $bddfont
    $c itemconfigure edge -width 2
    zoom $c 0.8
    insert "end generating dot graph" blue
   }
}

#
# invariant ...
#  check satifiability/validity of a fusion logic formula in 
#  terms of bdd operations 
#
proc  invariant {suffix_name_list negate_init_formula invariant_length Init Rho Tau} {
   variable a 
   variable b 
   variable a_cube 
   variable b_cube 
   variable manager
   variable my_names
   variable pair

   set num_vars [llength $suffix_name_list]
   set tt [expr $num_vars-$invariant_length]
   
   #insert "*** # of dependencies=$invariant_length, # of independent variables=$tt."
   set max_iterations 100
   
    set SuccW $Init
    
    if {$negate_init_formula} {
        insert "***Testing for validity with finite time."
        set SuccW [Not $SuccW]
    } else {
        insert "***Testing for satisfiability with finite time."
    }

    
    set SuccWStar $SuccW
    set OldSuccWStar [Not $SuccWStar]
    set Convergence 0
    set SuccWList {}
    #set FirstSuccW $SuccW
    set TempBDD [* $SuccW $Tau]
    
    #dot $Rho
    #dot $SuccW 
    set j 0
    
    while { ([* $SuccW $Tau] == [Zero]) && ($j < $max_iterations)  } {
     insert "***Starting iteration $j."
     check_node_stats
     lappend SuccWList $SuccW
     set SuccW [Swap [AndExists $a_cube $SuccW $Rho] $pair]
     
     set OldSuccWStar $SuccWStar
     set SuccWStar [+ $SuccWStar $SuccW]
        if { $OldSuccWStar == $SuccWStar } {
            set Convergence 1 
        } else {
          set Convergence 0
        }  
      #  dot $SuccW
     incr j
     if { $Convergence } { break }
    }
    if {$j==1} {
       set iterations "iteration"
    } else {
       set iterations "iterations"
    }
    insert "$j $iterations performed."
    if { [* $SuccW $Tau] == [Zero] } {
        if { ! $Convergence } {
          insert "***No convergence after $max_iterations iterations!"
        } else {
            if { ! $negate_init_formula } {
                insert  "Unsatisfiable with finite time." red
            } else {
                insert "Valid with finite time." green
            }
        }
    } else {
       # dot $Tau
       if { ! $negate_init_formula } {
         insert "Satisfiable with finite time." green
         set Counter "model"
       } else {
         insert "Not valid with finite time." red
         set Counter "counter-model"
       }
       insert "***Now start to find a $Counter."
       #insert "Length of SuccWList = [llength $SuccWList]" 
       set StateStringList {}
        if {[llength $suffix_name_list] == $invariant_length } {
            set number_of_states [expr $j+1]
            insert "No independent variables."
            if {$number_of_states == 1 } { 
                 set states "state"  
            } else {
                 set states "states"
            }
            insert "Can use any interval with exactly $number_of_states $states as a $Counter."
        } else {
            set SuccW [* $SuccW $Tau]
            while {$j>=0} {
                #insert "j=$j"
                set StateString ""
                set k $invariant_length
                #dot $SuccW
                while { $k < [llength $suffix_name_list] } {
                  append StateString "[lindex $suffix_name_list $k]="
                  if { [* $SuccW [lindex $a $k] ] == [Zero] } {
                      append StateString "0 "
                  } else {
                      set SuccW [* $SuccW [lindex $a $k]]
                      append StateString "1 "
                  }
                 #   dot $SuccW
                  incr k
                }
                if {$j !=0} {
#                    puts "SuccWList 0 = [lindex $SuccWList 0], FirstSuccW = $FirstSuccW"
                    set TempBDD [lindex $SuccWList end]
                    #remove last element from SuccWlist
                    set SuccWList [lreplace $SuccWList end end]
                    #dot $TempBDD
                    set SuccW [AndExists $b_cube $TempBDD  [* $Rho [Swap $SuccW $pair]]]
                }
                lappend StateStringList $StateString
                incr j -1
            } 
            set number_of_states [llength $StateStringList]
            if {$number_of_states == 1 } { 
                 set states "state"  
            } else {
                 set states "states"
            }
            insert "Here is a $Counter with $number_of_states $states:"
            for {set j [expr [llength $StateStringList] -1]} {$j>=0} {incr j -1} {
                set k [expr [llength $StateStringList] -$j-1]
                set z [lindex $StateStringList $j]
                set z1 [split $z " "]
                set z2 [lsort -dictionary $z1]
                insert "   State $k: "
                foreach z3 $z2 {
                  if { [string range $z3 0 5] == "AUTHON" } {
                     insert "  $z3" red
                  } else { 
                    if { [string range $z3 0 5] == "AUTHOP" } {
                     insert "  $z3" green
                    } else {
                      insert "  $z3"
                    }
                  }
                }   
            } 
        }
    }
    
    #print_bdd_stats
    insert "Buddy current number of live nodes=[bdd_getnodenum]" blue
    insert "----------------------------------------------"
}


#
# enforce ...
# enforce a fusion logic formula wrt  inputtrace in terms of bdd operations
#
proc  enforce {suffix_name_list invariant_length Init Rho Tau Inputtrace} {
   variable a 
   variable b 
   variable a_cube 
   variable b_cube 
   variable manager
   variable my_names
   variable pair
   

   set num_vars [llength $suffix_name_list]
   set tt [expr $num_vars-$invariant_length]
   
   #insert "*** # of dependencies=$invariant_length, # of independent variables=$tt."
   set max_iterations 100

    #set input_0 [lindex $Inputtrace 0]
    #puts "enf: input_0 : $input_0"  
    #RAnd input_0_and $input_0 
    
    set SuccW $Init 
    insert "***Testing enforcer against input trace."
    
    #set SuccWStar $SuccW
    #set OldSuccWStar [Not $SuccWStar]
    #set Convergence 0
    #set SuccWList {}
    #set FirstSuccW $SuccW
    #set TempBDD [* $SuccW $Tau]
    #dot $Rho
    #dot $SuccW 
    set j 0
    
    foreach input_element $Inputtrace {      
      # RAnd input_element_and $input_element 
      #puts "input_el: $input_element"
       
       set SuccW [* $SuccW $input_element]
       
       set SuccW [* $SuccW $Tau]

       if { [* $SuccW $Tau] == [Zero] } { 
         insert "not a final state" red 
       } else {
         insert "final state" green
       }

       set StateString ""
       set k $invariant_length
       while { $k < [llength $suffix_name_list] } {
          append StateString "[lindex $suffix_name_list $k]="
          if { [* $SuccW [lindex $a $k] ] == [Zero] } {
             append StateString "0 "
          } else {
             set SuccW [* $SuccW [lindex $a $k]]
             append StateString "1 "
          }
          incr k
       }
       set z1 [split $StateString " "]
       set z2 [lsort -dictionary $z1]
       insert "   State $j: "
       incr j
       foreach z3 $z2 {
         if { [string range $z3 0 5] == "AUTHON" } {
            insert "  $z3" red
         } else { 
            if { [string range $z3 0 5] == "AUTHOP" } {
               insert "  $z3" green
            } else {
               insert "  $z3"
            }
         }
       }   

     set SuccW [Swap [AndExists $a_cube $SuccW $Rho] $pair]

     
     # set SuccW $Init
    }
    insert "Buddy current number of live nodes=[bdd_getnodenum]" blue
    insert "----------------------------------------------"
}

#
# i_enforce ...
# interactive enforcement of a fusion logic formula in terms of bdd operations
#
proc  i_enforce {suffix_name_list invariant_length Init Rho Tau} {
   variable a 
   variable b 
   variable a_cube 
   variable b_cube 
   variable manager
   variable my_names
   variable indep
   variable pair

   set num_vars [llength $suffix_name_list]
   set tt [expr $num_vars-$invariant_length]
   
   #insert "*** # of dependencies=$invariant_length, # of independent variables=$tt."
   set max_iterations 100

    #set input_0 [lindex $Inputtrace 0]
    #puts "enf: input_0 : $input_0"  
    #RAnd input_0_and $input_0 
    
    set SuccW $Init 
    insert "***Testing enforcer against interactive input."
    
    #set SuccWStar $SuccW
    #set OldSuccWStar [Not $SuccWStar]
    #set Convergence 0
#    set SuccWList {}
    #set FirstSuccW $SuccW
    #set TempBDD [* $SuccW $Tau]
    #dot $Rho
    #dot $SuccW 
    set j 0
     
     set g_input ""
     while { $g_input != "end" } {
      puts "new input"
      
      gets stdin g_input
      update 

      if { $g_input != "end" } {
         #puts "g_input: $g_input"
         set g_input_a [subst $g_input]
         #puts "g_input_a: $g_input_a"
         set inp [tl_and $g_input_a]
         #puts "inp: $inp"
         set inp_a [input_to_trace $inp]
         #puts "inp_a: $inp_a"
         set inp_bdd [subst $inp_a]
       # puts "el_bdd: $el_bdd"
         # RAnd input_element_and $input_element 
         #puts "input_el: $input_element"
          
 
       set SuccW [* $SuccW $inp_bdd]
       
       set SuccW [* $SuccW $Tau]

       if { [* $SuccW $Tau] == [Zero] } { 
         puts "not a final state" 
       } else {
         puts "final state" 
       }

       set StateString ""
       set k $invariant_length
       while { $k < [llength $suffix_name_list] } {
          append StateString "[lindex $suffix_name_list $k]="
          if { [* $SuccW [lindex $a $k] ] == [Zero] } {
             append StateString "0 "
          } else {
             set SuccW [* $SuccW [lindex $a $k]]
             append StateString "1 "
          }
          incr k
          update
       }
       set z1 [split $StateString " "]
       set z2 [lsort -dictionary $z1]
       puts "   State $j: "
       incr j
       foreach z3 $z2 {
         if { [string range $z3 0 5] == "AUTHON" } {
            puts "  $z3" 
         } else { 
            if { [string range $z3 0 5] == "AUTHOP" } {
               puts "  $z3" 
            } else {
               puts "  $z3"
            }
         }
       }   

     set SuccW [Swap [AndExists $a_cube $SuccW $Rho] $pair]
     
     update 
     # set SuccW $Init
    insert "Buddy current number of live nodes=[bdd_getnodenum]" blue
    insert "----------------------------------------------"
      } else { 
        return
      }
     }
}

